@@ -127,11 +127,8 @@ radeon_get_scrninfo(int entity_num, void *pci_dev)
     if (!pScrn)
         return FALSE;
 
-    if (pci_dev) {
-      if (!radeon_kernel_mode_enabled(pScrn, pci_dev)) {
+    if (!radeon_kernel_mode_enabled(pScrn, pci_dev))
 	return FALSE;
-      }
-    }
 
     pScrn->driverVersion = RADEON_VERSION_CURRENT;
     pScrn->driverName    = RADEON_DRIVER_NAME;
@@ -154,7 +151,6 @@ radeon_get_scrninfo(int entity_num, void *pci_dev)
      */
     {
         DevUnion    *pPriv;
-        RADEONEntPtr pRADEONEnt;
 
         xf86SetEntitySharable(entity_num);
 
@@ -166,14 +162,8 @@ radeon_get_scrninfo(int entity_num, void *pci_dev)
 
 	xf86SetEntityInstanceForScreen(pScrn, pEnt->index, xf86GetNumEntityInstances(pEnt->index) - 1);
 
-        if (!pPriv->ptr) {
+        if (!pPriv->ptr)
             pPriv->ptr = xnfcalloc(sizeof(RADEONEntRec), 1);
-            pRADEONEnt = pPriv->ptr;
-            pRADEONEnt->HasSecondary = FALSE;
-        } else {
-            pRADEONEnt = pPriv->ptr;
-            pRADEONEnt->HasSecondary = TRUE;
-        }
     }
 
     free(pEnt);
@@ -202,6 +192,10 @@ RADEONDriverFunc(ScrnInfoPtr scrn, xorgDriverFuncOp op, void *data)
 	    flag = (CARD32 *)data;
 	    (*flag) = 0;
 	    return TRUE;
+#if XORG_VERSION_CURRENT > XORG_VERSION_NUMERIC(1,15,99,0,0)
+	case SUPPORTS_SERVER_FDS:
+	    return TRUE;
+#endif
 	default:
 	    return FALSE;
     }
@@ -267,11 +261,10 @@ radeon_platform_probe(DriverPtr pDriver,
         if (!pPriv->ptr) {
             pPriv->ptr = xnfcalloc(sizeof(RADEONEntRec), 1);
             pRADEONEnt = pPriv->ptr;
-            pRADEONEnt->HasSecondary = FALSE;
         } else {
             pRADEONEnt = pPriv->ptr;
-            pRADEONEnt->HasSecondary = TRUE;
         }
+        pRADEONEnt->platform_dev = dev;
     }
 
     free(pEnt);
